import Metal
import MetalPerformanceShaders

// Calculates matrix multiplication floating point operations
func getFops(matrixDim: Int) -> Int {
    return (2 * matrixDim - 1) * matrixDim * matrixDim 
}

// Reports the mean flops in units of Teraflops 
func getTflops(nFP: Int, time: Double) -> String {
    return String(format: "%.3f", 1e-12 * Double(nFP) / time)
} 

// Get the device, commandQueue, commandBuffer and blitEncoder
let device = MTLCreateSystemDefaultDevice()! 
let commandQueue = device.makeCommandQueue()!
let commandBuffer = commandQueue.makeCommandBuffer()!
let blitEncoder = commandBuffer.makeBlitCommandEncoder()!

// Matrix dimensions
let n = 8192
let rowsA = n
let colsA = n
let rowsB = n
let colsB = n
let rowsC = n
let colsC = n

// Set data for Matrix A, B 
let a = UnsafeMutablePointer<Float>.allocate(capacity: rowsA * colsA)
let arrayA = UnsafeMutableBufferPointer(start: a, count: rowsA * colsA)
arrayA.assign(repeating: Float(1.0))
print("Values in matrix A[\(rowsA) x \(colsA)]: \(arrayA[0]) uniformly")

let b = UnsafeMutablePointer<Float>.allocate(capacity: rowsA * colsA)
let arrayB = UnsafeMutableBufferPointer(start: b, count: rowsB * colsB)
arrayB.assign(repeating: Float(2.0))
print("Values in matrix B[\(rowsB) x \(colsB)]: \(arrayB[0]) uniformly")

// 1. Prepare managed buffers
// matrix A 
let rowBytesA = colsA * MemoryLayout<Float>.stride
let bufferA = device.makeBuffer(bytes: arrayA.baseAddress!,
                length: rowsA * rowBytesA, options: [.storageModeManaged])!
// matrix B
let rowBytesB = colsB * MemoryLayout<Float>.stride
let bufferB = device.makeBuffer(bytes: arrayB.baseAddress!,
                length: rowsB * rowBytesB, options: [.storageModeManaged])!
// matrix C
let rowBytesC = colsC * MemoryLayout<Float>.stride
let bufferC = device.makeBuffer(length: colsC * rowBytesC,
                options: [.storageModeManaged])!  
  
// 2. Encode matrix multiplication 
let descrA = MPSMatrixDescriptor(rows: rowsA, columns: colsA, rowBytes: rowBytesA, dataType: .float32) 
let descrB = MPSMatrixDescriptor(rows: rowsB, columns: colsB, rowBytes: rowBytesB, dataType: .float32) 
let descrC = MPSMatrixDescriptor(rows: rowsC, columns: colsC, rowBytes: rowBytesC, dataType: .float32) 

let matrixA = MPSMatrix(buffer: bufferA, descriptor: descrA) 
let matrixB = MPSMatrix(buffer: bufferB, descriptor: descrB) 
let matrixC = MPSMatrix(buffer: bufferC, descriptor: descrC)

print("Starting calculation on \(device.name)\n...")
let startTime = CFAbsoluteTimeGetCurrent()
let matMul = MPSMatrixMultiplication(device: device, resultRows: colsC, resultColumns: colsC, interiorColumns: colsB)
matMul.encode(commandBuffer: commandBuffer, leftMatrix: matrixA, rightMatrix: matrixB, resultMatrix: matrixC)
  
// 3. Get data back from GPU  
blitEncoder.synchronize(resource: bufferC)
  
// 4. Run buffer  
commandBuffer.commit()
commandBuffer.waitUntilCompleted()
let elapsedTime = Double(Int(1000 * (CFAbsoluteTimeGetCurrent() - startTime))) / 1000
  
// Read results
let nC = rowsC * colsC
let resultPointer = bufferC.contents().bindMemory(to: Float.self, capacity: nC)
let result = UnsafeBufferPointer(start: resultPointer, count: nC)

// Check consistency of resulting matrix
var ok = true
for i in 1..<nC {
    if result[i] != result[0] {
        ok = false
    }
}
if (ok) {
    print("Values in matrix C = A * B: \(result[0]) uniformly")
    let fops = getFops(matrixDim : n)
    let tFlops = getTflops(nFP: fops, time: elapsedTime)
    print("\(fops) floating point operations performed")
    print("Elapsed GPU time = \(elapsedTime) seconds -> \(tFlops) Teraflops")
}
else {
    print("Error: Inconsistent calculation results")
}
